/**************************************************************************
	MMPTP.CPP - Command line interface for MMPTP

	(C) 2002 by Michael Minn - michael@michaelminn.com
	(C) Copyright 2002 Jos�Luis S�chez, wizord@mi.madritel.es

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

****************************************************************************/

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include "mmptp.h"


/********************** mmptp_container **************************/

char *mmptp_container::read_string()
{
	int length = (int) current[0];

	// printf("%d byte string\n", length);

	char *string = new char[length + 1];
	for (int scan = 0; scan < length; ++scan)
		string[scan] = (char)current[1 + (scan * 2)];

	string[length] = 0;
	current += (1 + (length * 2));

	return string;
}

unsigned short *mmptp_container::read_uint16_array()
{
	int length = *((int*)current);

	if (length > 100)
	{
		fprintf(stderr, "Response container format error at position %d\n", current - payload);
		return 0;
	}

	// printf("%d entry array\n", length);

	unsigned short *array = new unsigned short[length + 1];
	memcpy(array, &current[4], length * 2);

	array[length] = 0; // null terminator
	current += (4 + (length * 2));

	return array;
}

void mmptp_container::hex_dump()
{
	char *buffer = (char*) this;

	for (int scan = 0; (scan < length) && (scan < 256); ++scan)
	{
		if ((buffer[scan] >= ' ') && (buffer[scan] <= '~'))
			printf("%02x: %02x %c\n", scan, (unsigned int) buffer[scan], buffer[scan]);
		else
			printf("%02x: %02x\n", scan, (unsigned int) buffer[scan]);
	}
}

int mmptp_container::transaction(int device)
{
	int status = 0;
	mmptp_container response;

	if ((status = write(device, this, length)) < (int) length)
	{
		fprintf(stderr, "Failure sending command to %s: %s\n", MMPTP_DEVICE_FILE, strerror(errno));
		status = -errno;
	}

	else if (reset() == NULL)
		fprintf(stderr, "reset() failure (can't occur)\n");

	else if ((status = read(device, this, sizeof(mmptp_container))) < 12)
	{
		fprintf(stderr, "Failure receiving response from %s: %s\n", MMPTP_DEVICE_FILE, strerror(errno));
		status = -errno;
	}

	else if (type == MMPTP_CONTAINER_TYPE_RESPONSE)
	{
		if (code != MMPTP_RESPONSE_OK)
		{
			fprintf(stderr, "Transaction failed: %s\n", response_code_string());
			status = -(int) code;
		}
		else
			status = code;
	}

	else if (type != MMPTP_CONTAINER_TYPE_DATA)
	{
		fprintf(stderr, "Invalid container type %d for data returned from device (status %d)\n", type, status);
		hex_dump();
		status = -EBADMSG;
	}

	else if ((status = read(device, &response, sizeof(mmptp_container))) < 12)
	{
		fprintf(stderr, "Failure receiving response from %s: %s\n", MMPTP_DEVICE_FILE, strerror(errno));
		status = -errno;
	}

	else if (response.type != MMPTP_CONTAINER_TYPE_RESPONSE)
	{
		fprintf(stderr, "Invalid container type %d response from device\n", response.type);
		response.hex_dump();
		status = -EBADMSG;
	}

	else if (response.code != MMPTP_RESPONSE_OK)
	{
		fprintf(stderr, "Transaction failed: %s\n", response_code_string());
		status = -(int)response.code;
	}

	else
		status = code;

	return status;
}

char *mmptp_container::response_code_string()
{
	switch (code)
	{
		case MMPTP_RESPONSE_OK: return "OK";
		case MMPTP_RESPONSE_GENERAL_ERROR: return "General Error";
		case MMPTP_RESPONSE_SESSION_NOT_OPEN: return "Session Not Open";
		case MMPTP_RESPONSE_INVALID_TRANSACTION_ID: return "Invalid Transaction ID";
		case MMPTP_RESPONSE_OPERATION_NOT_SUPPORTED: return "Operation Not Supported";
		case MMPTP_RESPONSE_PARAMETER_NOT_SUPPORTED: return "Parameter Not Supported";
		case MMPTP_RESPONSE_INCOMPLETE_TRANSFER: return "Incomplete Transfer";
		case MMPTP_RESPONSE_INVALID_STORAGE_ID: return "Invalid Storage ID";
		case MMPTP_RESPONSE_INVALID_OBJECTHANDLE: return "Invalid Object Handle";
		case MMPTP_RESPONSE_DEVICEPROP_NOT_SUPPORTED: return "Device Property Not Supported";
		case MMPTP_RESPONSE_INVALID_OBJECTFORMAT_CODE: return "Invalid Object Format Code";
		case MMPTP_RESPONSE_STORE_FULL: return "Store Full";
		case MMPTP_RESPONSE_OBJECT_WRITE_PROTECTED: return "Object Write Protected";
		case MMPTP_RESPONSE_STORE_READ_ONLY: return "Store Is Read Only";
		case MMPTP_RESPONSE_ACCESS_DENIED: return "Access Denied";
		case MMPTP_RESPONSE_NO_THUMBNAIL_PRESENT: return "No Thumbnail Present";
		case MMPTP_RESPONSE_SELFTEST_FAILED: return "Self Test Failed";
		case MMPTP_RESPONSE_PARTIAL_DELETION: return "Partial Deletion";
		case MMPTP_RESPONSE_STORE_NOT_AVAILABLE: return "Store Not Available";
		case MMPTP_RESPONSE_SPECIFICATION_UNSUPPORTED: return "Specification By Format Unsupported";
		case MMPTP_RESPONSE_NO_VALID_OBJECTINFO: return "No Valid Object Info";
		case MMPTP_RESPONSE_INVALID_CODE_FORMAT: return "Invalid Code Format";
		case MMPTP_RESPONSE_UNKNOWN_VENDOR_CODE: return "Unknown Vendor Code";
		case MMPTP_RESPONSE_CAPTURE_ALREADY_TERMINATED: return "Capture Already Terminated";
		case MMPTP_RESPONSE_DEVICE_BUSY: return "Device Busy";
		case MMPTP_RESPONSE_INVALID_PARENTOBJECT: return "Invalid Parent Object";
		case MMPTP_RESPONSE_INVALID_DEVICEPROP_FORMAT: return "Invalid Device Property Format";
		case MMPTP_RESPONSE_INVALID_DEVICEPROP_VALUE: return "Invalid Device Property Value";
		case MMPTP_RESPONSE_INVALID_PARAMETER: return "Invalid Parameter";
		case MMPTP_RESPONSE_SESSION_ALREADY_OPEN: return "Session Already Open";
		case MMPTP_RESPONSE_TRANSACTION_CANCELLED: return "Transaction Cancelled";
		case MMPTP_RESPONSE_DESTINATION_UNSUPPORTED: return "Destination Unsupported";
	}

	static char default_message[32];
	sprintf(default_message, "Invalid Response %x", code);
	return default_message;
}

int mmptp_container::open_session(int device)
{
	length = 16;
	type = MMPTP_CONTAINER_TYPE_COMMAND;
	code = MMPTP_OPERATION_OPENSESSION;
	transaction_id = 0;
	write_uint32(time(NULL));

	return (status = transaction(device));
}

int mmptp_container::get_storage_ids(int device, unsigned int *(&storage_ids))
{
	storage_ids = NULL;

	length = 12;
	type = MMPTP_CONTAINER_TYPE_COMMAND;
	code = MMPTP_OPERATION_GETSTORAGEIDS;
	++transaction_id;

	if ((status = transaction(device)) < 0)
		return status;

	reset();
	unsigned int count = read_uint32();

	if ((storage_ids = new unsigned int[count]) == NULL)
	{
		fprintf(stderr, "Failure allocating %d byte Storage ID array: %s\n", count, strerror(errno));
		return -errno;
	}

	for (status = 0; status < count; ++status)
		storage_ids[status] = read_uint32();

	return status;
}

int mmptp_container::get_storage_info(int device, unsigned int storage_id, mmptp_storageinfo &storage_info)
{
	length = 16;
	type = MMPTP_CONTAINER_TYPE_COMMAND;
	code = MMPTP_OPERATION_GETSTORAGEINFO;
	++transaction_id;
	reset();
	write_uint32(storage_id);

	if ((status = transaction(device)) < 0)
		fprintf(stderr, "Failure getting storage id %x: %d %s\n", storage_id, status, strerror(errno));
	else
	{
		storage_info.copy_contents(this);
		status = sizeof(storage_info);
	}

	return status;
}

int mmptp_container::get_object_handles(int device, unsigned int storage_id, unsigned int *(&handles))
{
printf("Getting handles for storage %x\n", storage_id);

	length = 16;
	type = MMPTP_CONTAINER_TYPE_COMMAND;
	code = MMPTP_OPERATION_GETOBJECTHANDLES;
	++transaction_id;
	reset();
	write_uint32(storage_id);

	if ((status = transaction(device)) < 0)
	{
		fprintf(stderr, "Failure getting object handles for storage %x: %s\n", storage_id, strerror(errno));
		return status;
	}

	status = read_uint32();
	handles = new unsigned int[status];
	memcpy(handles, payload + 4, status * 4);

	return status;
}

int mmptp_container::get_object_info(int device, unsigned int object_handle, mmptp_objectinfo &objectinfo)
{
	length = 16;
	type = MMPTP_CONTAINER_TYPE_COMMAND;
	code = MMPTP_OPERATION_GETOBJECTINFO;
	++transaction_id;
	reset();
	write_uint32(object_handle);

	if ((status = transaction(device)) < 0)
		return status;

	objectinfo.copy_contents(this);

	return sizeof(*this);
}

int mmptp_container::get_object(int device, unsigned int object_handle)
{
	length = 16;
	type = MMPTP_CONTAINER_TYPE_COMMAND;
	code = MMPTP_OPERATION_GETOBJECT;
	++transaction_id;
	reset();
	write_uint32(object_handle);

	return (status = transaction(device));
}

int mmptp_container::close_session(int device)
{
printf("closing session...\n");
	length = 12;
	type = MMPTP_CONTAINER_TYPE_COMMAND;
	code = MMPTP_OPERATION_CLOSESESSION;
	++transaction_id;

	return (status = transaction(device));
}


/****** Description structure for info returned about devices *******/

mmptp_deviceinfo::~mmptp_deviceinfo()
{
	if (vendor_extension_description)
		delete vendor_extension_description;
	if (operations_supported)
		delete operations_supported;
	if (events_supported)
		delete events_supported;
	if (device_properties_supported)
		delete device_properties_supported;
	if (capture_formats)
		delete capture_formats;
	if (image_formats)
		delete image_formats;
	if (manufacturer)
		delete manufacturer;
	if (model)
		delete model;
	if (device_version)
		delete device_version;
	if (serial_number)
		delete serial_number;

	memset(this, 0, sizeof(*this));
}

int mmptp_deviceinfo::copy_contents(mmptp_container *container)
{
	container->reset();

	standard_version = container->read_uint16();
	vendor_extension_id = container->read_uint32();
	vendor_extension_version = container->read_uint16();
	vendor_extension_description = container->read_string();
	functional_mode = container->read_uint16();
	operations_supported = container->read_uint16_array();
	events_supported = container->read_uint16_array();
	device_properties_supported = container->read_uint16_array();
	capture_formats = container->read_uint16_array();
	image_formats = container->read_uint16_array();
	manufacturer = container->read_string();
	model = container->read_string();
	device_version = container->read_string();
	serial_number = container->read_string();

	return sizeof(*this);
}

int mmptp_deviceinfo::list_contents()
{
	int scan;

	printf("Standard Version: %f\n", standard_version / 100.0);
	printf("Vendor Extension: ID %d\n", vendor_extension_id);
	printf("Vendor Extension Version: %f\n", vendor_extension_version / 100.0);
	printf("Vendor Extension Description: %s\n", vendor_extension_description);
	printf("Functional Mode %d\n", functional_mode);

	printf("Operations Supported: ");
	for (scan = 0; operations_supported[scan]; ++scan)
		printf("%x ", operations_supported[scan]);
	printf("\n");

	printf("Events Supported: ");
	for (scan = 0; events_supported[scan]; ++scan)
		printf("%x ", events_supported[scan]);
	printf("\n");

	printf("Device Properties Supported: ");
	for (scan = 0; device_properties_supported[scan]; ++scan)
		printf("%x ", device_properties_supported[scan]);
	printf("\n");

	printf("Capture Formats Supported: ");
	for (scan = 0; capture_formats[scan]; ++scan)
		printf("%x ", capture_formats[scan]);
	printf("\n");

	printf("Image Formats Supported: ");
	for (scan = 0; image_formats[scan]; ++scan)
		printf("%x ", image_formats[scan]);
	printf("\n");

	printf("Manufacturer: %s\n", manufacturer);
	printf("Model: %s\n", model);
	printf("Device Version: %s\n", device_version);
	printf("Serial Number: %s\n", serial_number);

	return 0;
}

/****** Description structure for info returned about storage areas on devices *******/

mmptp_storageinfo::~mmptp_storageinfo()
{
	if (storage_description)
		delete storage_description;
	if (volume_label)
		delete volume_label;

	memset(this, 0, sizeof(*this));
}

int mmptp_storageinfo::copy_contents(mmptp_container *container)
{
	container->reset();

	storage_type = container->read_uint16();
	filesystem_type = container->read_uint16();
	access_capability = container->read_uint16();
	max_capacity_lsw = container->read_uint32();
	max_capacity_msw = container->read_uint32();
	free_space_bytes_lsw = container->read_uint32();
	free_space_bytes_msw = container->read_uint32();
	free_images = container->read_uint32();
	storage_description = container->read_string();
	volume_label = container->read_string();

	return sizeof(*this);
}

int mmptp_storageinfo::list_contents()
{
	int scan;

	printf("Storage type %x\n", storage_type);
	printf("File system type %x\n", filesystem_type);
	printf("Access capability %x\n", access_capability);
	printf("Capacity %f\n", (max_capacity_lsw * 1.0) + (max_capacity_msw * 65536.0));
	printf("Free space %f\n", (free_space_bytes_lsw * 1.0) + (free_space_bytes_msw * 65536.0));
	printf("Free images %d\n", free_images);
	printf("Description: %s\n", storage_description);
	printf("Volume label: %s\n", volume_label);

	return 0;
}


/****** Description structure for info returned about objects stored on devices *******/

mmptp_objectinfo::~mmptp_objectinfo()
{
	if (file_name)
		delete file_name;
	if (capture_date)
		delete capture_date;
	if (modification_date)
		delete modification_date;
	if (keywords)
		delete keywords;

	memset(this, 0, sizeof(*this));
}

int mmptp_objectinfo::copy_contents(mmptp_container *container)
{
	container->reset();

	storage_id = container->read_uint32();
	object_format = container->read_uint16();
	protection_status = container->read_uint16();
	object_compressed_size = container->read_uint32();
	thumb_format = container->read_uint16();
	thumb_compressed_size = container->read_uint32();
	thumb_pixel_width = container->read_uint32();
	thumb_pixel_height = container->read_uint32();
	image_pixel_width = container->read_uint32();
	image_pixel_height = container->read_uint32();
	image_bit_depth = container->read_uint32();
	parent_object = container->read_uint32();
	association_type = container->read_uint16();
	association_description = container->read_uint32();
	sequence_number = container->read_uint32();
	file_name = container->read_string();
	capture_date = container->read_string();
	modification_date = container->read_string();
	keywords = container->read_string();

	return sizeof(*this);
}

int mmptp_objectinfo::list_contents()
{
	printf("Storage ID: %x\n", storage_id);
	printf("Format: %x\n", object_format);
	printf("Protection: %x\n", protection_status);
	printf("Size: %d\n", object_compressed_size);
	printf("Thumbnail format: %x\n", thumb_format);
	printf("Thumbnail size: %d\n", thumb_compressed_size);
	printf("Thumbnail width: %d\n", thumb_pixel_width);
	printf("Thumbnail height: %d\n", thumb_pixel_height);
	printf("Width: %d\n", image_pixel_width);
	printf("Height: %d\n", image_pixel_height);
	printf("Depth: %d\n", image_bit_depth);
	printf("Parent: %x\n", parent_object);
	printf("Association: %x\n", association_type);
	printf("Asso. Description: %x\n", association_description);
	printf("Sequence #: %d\n", sequence_number);
	printf("File name: %s\n", file_name);
	printf("Capture date: %s\n", capture_date);
	printf("Modification date: %s\n", modification_date);
	printf("Keywords: %s\n\n", keywords);

	return 0;
}

int mmptp_objectinfo::list_contents_short()
{
	printf("%s    %4.4s/%2.2s/%2.2s %2.2s:%2.2s:%2.2s    %dx%d    (%d bytes)\n",
		file_name,
		&capture_date[0], &capture_date[4],
		&capture_date[6], &capture_date[9],
		&capture_date[11], &capture_date[13],
		image_pixel_width, image_pixel_height,
		object_compressed_size);


	return 0;
}



/************************* Program main() entry function ******************************/

int print_deviceinfo()
{
	int status = 0;
	int device = open(MMPTP_DEVICE_FILE, O_RDWR);

	if (device < 0)
	{
		fprintf(stderr, "Failure opening MMPTP device %s: %s\n", MMPTP_DEVICE_FILE, strerror(errno));
		return 0;
	}

	struct mmptp_container command;
	command.length = 12;
	command.type = MMPTP_CONTAINER_TYPE_COMMAND;
	command.code = MMPTP_OPERATION_GETDEVICEINFO;
	command.transaction_id = 0;

	if ((status = command.transaction(device)) < 0)
		return status;

	mmptp_deviceinfo deviceinfo;
	deviceinfo.copy_contents(&command);
	deviceinfo.list_contents();

	close (device);
	return 0;
}


int print_storageinfo()
{
	int id_count = 0;
	int device = open(MMPTP_DEVICE_FILE, O_RDWR);
	unsigned int *storage_ids = NULL;
	mmptp_container container;
	mmptp_storageinfo storage_info;

	if (device < 0)
	{
		fprintf(stderr, "Failure opening MMPTP device %s: %s\n", MMPTP_DEVICE_FILE, strerror(errno));
		return 0;
	}

	if (container.open_session(device) < 0)
		return container.status;

	else if ((id_count = container.get_storage_ids(device, storage_ids)) < 0)
	{
		container.close_session(device);
		return container.status;
	}

	else
	{
		printf("%d Storage units on device --------------\n", id_count);
		for (int id = 0; id < id_count; ++id)
			if ((storage_ids[id] & 0x00ff) == 0)
				printf("\nStorage ID %x\nNot Connected\n", storage_ids[id]);
			else if (container.get_storage_info(device, storage_ids[id], storage_info) >= 0)
			{
				printf("\nStorage ID %x\n", storage_ids[id]);
				storage_info.list_contents();
			}
	}

	container.close_session(device);

	close (device);

	return 0;
}

int list_objects(int long_format)
{
	int device = open(MMPTP_DEVICE_FILE, O_RDWR);
	int transaction = 1;
	int storage_count;
	int object_count = 0;
	int retval = 0;
	unsigned int *storage_ids = NULL;
	unsigned int *handles = NULL;
	mmptp_container container;
	mmptp_objectinfo objectinfo;

	if (device < 0)
	{
		fprintf(stderr, "Failure opening MMPTP device %s: %s\n", MMPTP_DEVICE_FILE, strerror(errno));
		return 0;
	}

	if (container.open_session(device) < 0)
		return container.status;

	else if ((storage_count = container.get_storage_ids(device, storage_ids)) < 0)
	{
		container.close_session(device);
		return container.status;
	}

	else for (int id = 0; id < storage_count; ++id)
		if ((storage_ids[id] & 0x00ff) == 0)
			printf("\nStorage ID %x\nNot Connected\n", storage_ids[id]);

		else if ((object_count = container.get_object_handles(device, storage_ids[id], handles)) < 0)
			return container.status;

		else if (object_count == 0)
			printf("No objects on device\n");

		else for (int scan = 0; scan < object_count; ++scan)
			if (container.get_object_info(device, handles[scan], objectinfo) < 0)
				return container.status;
			else if (long_format)
				objectinfo.list_contents();
			else
				objectinfo.list_contents_short();

	container.close_session(device);

	close (device);

	return retval;
}

int download_objects()
{
	int device = open(MMPTP_DEVICE_FILE, O_RDWR);
	int transaction = 1;
	int storage_count;
	int object_count = 0;
	int retval = 0;
	unsigned int *storage_ids = NULL;
	unsigned int *handles = NULL;
	mmptp_container container;
	mmptp_objectinfo objectinfo;

	if (device < 0)
	{
		fprintf(stderr, "Failure opening MMPTP device %s: %s\n", MMPTP_DEVICE_FILE, strerror(errno));
		return 0;
	}

	if (container.open_session(device) < 0)
		return container.status;

	else if ((storage_count = container.get_storage_ids(device, storage_ids)) < 0)
	{
		container.close_session(device);
		return container.status;
	}

	else for (int id = 0; id < storage_count && retval == 0; ++id)
		if ((storage_ids[id] & 0x00ff) == 0)
			printf("\nStorage ID %x\nNot Connected\n", storage_ids[id]);

		else if ((object_count = container.get_object_handles(device, storage_ids[id], handles)) < 0)
			retval = container.status;

		else if (object_count == 0)
			printf("No objects on device\n");

		else for (int scan = 0; scan < object_count && retval == 0; ++scan)
			if (container.get_object_info(device, handles[scan], objectinfo) < 0)
				retval = container.status;

			else if (objectinfo.object_format == MMPTP_FORMAT_ASSOCIATION)
				; // ignore directory objects

			else if (container.get_object(device, handles[scan]) < 0)
				retval = container.status;

			else
			{
				char filename[256];
				if (strlen(objectinfo.capture_date) >= 15)
					sprintf(filename, "%4.4s-%2.2s-%2.2s_%2.2s-%2.2s-%2.2s.jpg",
					&objectinfo.capture_date[0], &objectinfo.capture_date[4],
					&objectinfo.capture_date[6], &objectinfo.capture_date[9],
					&objectinfo.capture_date[11], &objectinfo.capture_date[13]);
				else
					sprintf(filename, "%s.jpg", objectinfo.capture_date);

				printf("Photo %x: %s (%d bytes)\n", handles[scan], filename, container.length);

				int photofile = open(filename, O_RDWR | O_CREAT, 00666);
				if (photofile < 0)
					fprintf(stderr, "Failure opening output file %s: %s\n", filename, strerror(errno));

				else if (write(photofile, container.payload, container.length) < container.length)
					fprintf(stderr, "Failure writing %d byte output file %s: %s\n", container.length, filename, strerror(errno));

				if (photofile >= 0)
					close(photofile);
			}

	container.close_session(device);

	close (device);

	return retval;
}

int print_syntax()
{
	printf("SYNTAX: mmptp -i | -s | -l | -L | -d\n");
	printf("        -i: print device info\n");
	printf("        -s: print storage info\n");
	printf("        -l: list images (short format)\n");
	printf("        -L: list images (long format)\n");
	printf("        -d: download images to current directory\n");
	return 0;
}


int main(int argc, char **argv)
{
	printf("MMPTPCMD v2003-07-19 by Michael Minn\n");

	if (argc != 2)
		return print_syntax();

	else if (strcmp(argv[1], "-i") == 0)
		return print_deviceinfo();

	else if (strcmp(argv[1], "-s") == 0)
		return print_storageinfo();

	else if (strcmp(argv[1], "-l") == 0)
		return list_objects(0);

	else if (strcmp(argv[1], "-L") == 0)
		return list_objects(1);

	else if (strcmp(argv[1], "-d") == 0)
		return download_objects();

	else
		return print_syntax();
}



