/**************************************************************************
	MMPTP - Linux USB Driver for PTP digital cameras

	(C) 2002 by Michael Minn - michael@michaelminn.com

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

****************************************************************************/

#ifndef MMPTP_H
#define MMPTP_H

#define MMPTP_CONTAINER_TYPE_UNDEFINED 0
#define MMPTP_CONTAINER_TYPE_COMMAND 1
#define MMPTP_CONTAINER_TYPE_DATA 2
#define MMPTP_CONTAINER_TYPE_RESPONSE 3
#define MMPTP_CONTAINER_TYPE_EVENT 4


/* Operation Codes - PIMA 15740:2000 spec, clause 10.4, pp 61 */

#define MMPTP_OPERATION_GETDEVICEINFO 0x1001
#define MMPTP_OPERATION_OPENSESSION 0x1002
#define MMPTP_OPERATION_CLOSESESSION 0x1003
#define MMPTP_OPERATION_GETSTORAGEIDS 0x1004
#define MMPTP_OPERATION_GETSTORAGEINFO 0x1005
#define MMPTP_OPERATION_GETNUMOBJECTS 0x1006
#define MMPTP_OPERATION_GETOBJECTHANDLES 0x1007
#define MMPTP_OPERATION_GETOBJECTINFO 0x1008
#define MMPTP_OPERATION_GETOBJECT 0x1009
#define MMPTP_OPERATION_GETTHUMB 0x100a
#define MMPTP_OPERATION_DELETEOBJECT 0x100b
#define MMPTP_OPERATION_SENDOBJECTINFO 0x100c
#define MMPTP_OPERATION_SENDOBJECT 0x100d
#define MMPTP_OPERATION_INITIATECAPTURE 0x100e
#define MMPTP_OPERATION_FORMATSTORE 0x100f
#define MMPTP_OPERATION_RESETDEVICE 0x1010
#define MMPTP_OPERATION_SELFTEST 0x1011
#define MMPTP_OPERATION_SETOJBJECTPROTECTION 0x1012
#define MMPTP_OPERATION_POWERDOWN 0x1013
#define MMPTP_OPERATION_GETDEVICEPROPDESC 0x1014
#define MMPTP_OPERATION_GETDEVICEPROPVALUE 0x1015
#define MMPTP_OPERATION_SETDEVICEPROPVALUE 0x1016
#define MMPTP_OPERATION_RESETDEVICEPROPVALUE 0x1017
#define MMPTP_OPERATION_TERMINATEOPENCAPTURE 0x1018
#define MMPTP_OPERATION_MOVEOBJECT 0x1019
#define MMPTP_OPERATION_COPYOBJECT 0x101a
#define MMPTP_OPERATION_GETPARTIALOBJECT 0x101b
#define MMPTP_OPERATION_INITIATEOPENCAPTURE 0x101c


/* Response Codes - PIMA 15740:2000 spec, clause 11, pp 85 */

#define MMPTP_RESPONSE_OK 0x2001
#define MMPTP_RESPONSE_GENERAL_ERROR 0x2002
#define MMPTP_RESPONSE_SESSION_NOT_OPEN 0x2003
#define MMPTP_RESPONSE_INVALID_TRANSACTION_ID 0x2004
#define MMPTP_RESPONSE_OPERATION_NOT_SUPPORTED 0x2005
#define MMPTP_RESPONSE_PARAMETER_NOT_SUPPORTED 0x2006
#define MMPTP_RESPONSE_INCOMPLETE_TRANSFER 0x2007
#define MMPTP_RESPONSE_INVALID_STORAGE_ID 0x2008
#define MMPTP_RESPONSE_INVALID_OBJECTHANDLE 0x2009
#define MMPTP_RESPONSE_DEVICEPROP_NOT_SUPPORTED 0x200a
#define MMPTP_RESPONSE_INVALID_OBJECTFORMAT_CODE 0x200b
#define MMPTP_RESPONSE_STORE_FULL 0x200c
#define MMPTP_RESPONSE_OBJECT_WRITE_PROTECTED 0x200d
#define MMPTP_RESPONSE_STORE_READ_ONLY 0x200e
#define MMPTP_RESPONSE_ACCESS_DENIED 0x200f
#define MMPTP_RESPONSE_NO_THUMBNAIL_PRESENT 0x2010
#define MMPTP_RESPONSE_SELFTEST_FAILED 0x2011
#define MMPTP_RESPONSE_PARTIAL_DELETION 0x2012
#define MMPTP_RESPONSE_STORE_NOT_AVAILABLE 0x2013
#define MMPTP_RESPONSE_SPECIFICATION_UNSUPPORTED 0x2014
#define MMPTP_RESPONSE_NO_VALID_OBJECTINFO 0x2015
#define MMPTP_RESPONSE_INVALID_CODE_FORMAT 0x2016
#define MMPTP_RESPONSE_UNKNOWN_VENDOR_CODE 0x2017
#define MMPTP_RESPONSE_CAPTURE_ALREADY_TERMINATED 0x2018
#define MMPTP_RESPONSE_DEVICE_BUSY 0x2019
#define MMPTP_RESPONSE_INVALID_PARENTOBJECT 0x201a
#define MMPTP_RESPONSE_INVALID_DEVICEPROP_FORMAT 0x201b
#define MMPTP_RESPONSE_INVALID_DEVICEPROP_VALUE 0x201c
#define MMPTP_RESPONSE_INVALID_PARAMETER 0x201d
#define MMPTP_RESPONSE_SESSION_ALREADY_OPEN 0x201e
#define MMPTP_RESPONSE_TRANSACTION_CANCELLED 0x201f
#define MMPTP_RESPONSE_DESTINATION_UNSUPPORTED 0x2020


/* Image Formats - PIMA 15740:2000 spec, clause 6.2, pp 37 */

#define MMPTP_FORMAT_ASSOCIATION 0x3001 /* i.e. folder */
#define MMPTP_FORMAT_SCRIPT 0x3002
#define MMPTP_FORMAT_EXECUTABLE 0x3003
#define MMPTP_FORMAT_TEXT 0x3004
#define MMPTP_FORMAT_HTML 0x3005
#define MMPTP_FORMAT_DPOF 0x3006
#define MMPTP_FORMAT_AIFF 0x3007
#define MMPTP_FORMAT_WAV 0x3008
#define MMPTP_FORMAT_MP3 0x3009
#define MMPTP_FORMAT_AVI 0x300a
#define MMPTP_FORMAT_MPEG 0x300b
#define MMPTP_FORMAT_ASF 0x300c
#define MMPTP_FORMAT_EXIFJPEG 0x3801
#define MMPTP_FORMAT_TIFF_EP 0x3802
#define MMPTP_FORMAT_FLASHPIX 0x3803
#define MMPTP_FORMAT_BMP 0x3804
#define MMPTP_FORMAT_CIFF 0x3805
#define MMPTP_FORMAT_GIF 0x3807
#define MMPTP_FORMAT_JPEG 0x3808
#define MMPTP_FORMAT_PCD 0x3809
#define MMPTP_FORMAT_PICT 0x380a
#define MMPTP_FORMAT_PNG 0x380b
#define MMPTP_FORMAT_TIFF 0x380d
#define MMPTP_FORMAT_TIFFIT 0x380e
#define MMPTP_FORMAT_JPEG2000 0x380f
#define MMPTP_FORMAT_JPEG2000X 0x3810


/* Event codes - PIMA 15740:2000 spec, clause 12.4, pp93 */

#define MMPTP_EVENT_CANCELTRANSACTION 0x4001
#define MMPTP_EVENT_OBJECTADDED 0x4002
#define MMPTP_EVENT_OBJECTREMOVED 0x4003
#define MMPTP_EVENT_STOREADDED 0x4004
#define MMPTP_EVENT_STOREREMOVED 0x4005
#define MMPTP_EVENT_DEVICEPROPCHANGED 0x4006
#define MMPTP_EVENT_OBJECTINFOCHANGED 0x4007
#define MMPTP_EVENT_DEVICEINFOCHANGED 0x4008
#define MMPTP_EVENT_REQUESTOBJECTTRANSFER 0x4009
#define MMPTP_EVENT_STOREFULL 0x400a
#define MMPTP_EVENT_DEVICERESET 0x400b
#define MMPTP_EVENT_STORAGEINFOCHANGED 0x400c
#define MMPTP_EVENT_CAPTURECOMPLETE 0x400d
#define MMPTP_EVENT_UNREPORTEDSTATUS 0x400e


/* Image Property Codes - PIMA 15740:2000 spec, clause 13.3.5, pp29 */

#define MMPTP_PROPERTY_BATTERYLEVEL 0x5001
#define MMPTP_PROPERTY_FUNCTIONALMODE 0x5002
#define MMPTP_PROPERTY_IMAGESIZE 0x5003
#define MMPTP_PROPERTY_COMPRESSIONSETTING 0x5004
#define MMPTP_PROPERTY_WHITEBALANCE 0x5005
#define MMPTP_PROPERTY_RGBGAIN 0x5006
#define MMPTP_PROPERTY_FNUMBER 0x5007
#define MMPTP_PROPERTY_FOCALLENGTH 0x5008
#define MMPTP_PROPERTY_FOCUSDISTANCE 0x5009
#define MMPTP_PROPERTY_FOCUSMODE 0x500a
#define MMPTP_PROPERTY_EXPOSUREMETERINGMODE 0x500b
#define MMPTP_PROPERTY_FLASHMODE 0x500c
#define MMPTP_PROPERTY_EXPOSURETIME 0x500d
#define MMPTP_PROPERTY_EXPOSUREPROGRAMMODE 0x500e
#define MMPTP_PROPERTY_EXPOSUREINDEX 0x500f
#define MMPTP_PROPERTY_EXPOSUREBIASCOMPENSATION 0x5010
#define MMPTP_PROPERTY_DATETIME 0x5011
#define MMPTP_PROPERTY_CAPTUREDELAY 0x5012
#define MMPTP_PROPERTY_STILLCAPTUREMODE 0x5013
#define MMPTP_PROPERTY_CONTRAST 0x5014
#define MMPTP_PROPERTY_SHARPNESS 0x5015
#define MMPTP_PROPERTY_DIGITALZOOM 0x5016
#define MMPTP_PROPERTY_EFFECTMODE 0x5017
#define MMPTP_PROPERTY_BURSTNUMBER 0x5018
#define MMPTP_PROPERTY_BURSTINTERVAL 0x5019
#define MMPTP_PROPERTY_TIMELAPSENUMBER 0x501a
#define MMPTP_PROPERTY_TIMELAPSEINTERVAL 0x501b
#define MMPTP_PROPERTY_FOCUSMETERINGMODE 0x501c
#define MMPTP_PROPERTY_UPLOADURL 0x501d
#define MMPTP_PROPERTY_ARTIST 0x501e
#define MMPTP_PROPERTY_COPYRIGHTINFO 0x501f


#ifndef MODULE

/************
Commands are sent to USB device in containers.
Data is returned from USB device in containers.
Containers longer than maximum urb packet length are sent in multiple successive packets
Generic container structures - PIMA 15740:2000 spec, clause D 7.1.1, pp 148
*************/

class mmptp_objectinfo;

class mmptp_storageinfo;

class mmptp_container
{
	public: // 1st 5 fields must be in this order
	unsigned int length;
	unsigned short type; // MMPTP_CONTAINER_TYPE_*
	unsigned short code; // operation, response or event code
	unsigned int transaction_id;
	unsigned char payload[2097152]; // arbitrary 2 meg buffer

	unsigned char *current;
	int status;

	mmptp_container() { memset(this, 0, sizeof(mmptp_container)); reset(); }
	int end_of_container() { return (current >= &payload[length - 12]); }
	char *reset() { current = (unsigned char*) payload; return (char*) current; }
	unsigned short read_uint16() { unsigned short value = *((unsigned short*)current); current += 2; return value; }
	unsigned int read_uint32() { unsigned int value = *((unsigned int*)current); current += 4; return value; }
	void write_uint16(unsigned short value) { *((unsigned short*)current) = value; current += 2; }
	void write_uint32(unsigned int value) { *((unsigned int*)current) = value; current += 4; }
	char *read_string();
	unsigned short *read_uint16_array();
	void hex_dump();

	int transaction(int device);
	char *response_code_string();
	int open_session(int device);
	int get_storage_ids(int device, unsigned int *(&storage_ids));
	int get_storage_info(int device, unsigned int storage_id, mmptp_storageinfo &storage_info);
	int get_object_handles(int device, unsigned int storage_id, unsigned int *(&handles));
	int get_object_info(int device, unsigned int object_handle, mmptp_objectinfo &objectinfo);
	int get_object(int device, unsigned int object_handle);
	int close_session(int device);
};

/* Response to getdeviceinfo - clause 5.5.1 pp 27 */

class mmptp_deviceinfo
{
	public:
	mmptp_deviceinfo() { memset(this, 0, sizeof(*this)); }
	~mmptp_deviceinfo();
	int copy_contents(mmptp_container *container);
	int list_contents();

	unsigned short standard_version;
	unsigned int vendor_extension_id;
	unsigned short vendor_extension_version;
	char *vendor_extension_description;
	unsigned short functional_mode;
	unsigned short *operations_supported;
	unsigned short *events_supported;
	unsigned short *device_properties_supported;
	unsigned short *capture_formats;
	unsigned short *image_formats;
	char *manufacturer;
	char *model;
	char *device_version;
	char *serial_number;
};

/* Response to getobjectinfo - clause 5.5.2 pp 30 */

class mmptp_objectinfo
{
	public:
	mmptp_objectinfo() { memset(this, 0, sizeof(*this)); }
	~mmptp_objectinfo();
	int copy_contents(mmptp_container *container);
	int list_contents();
	int list_contents_short();

	unsigned int storage_id;
	unsigned short object_format;
	unsigned short protection_status;
	unsigned int object_compressed_size;
	unsigned short thumb_format;
	unsigned int thumb_compressed_size;
	unsigned int thumb_pixel_width;
	unsigned int thumb_pixel_height;
	unsigned int image_pixel_width;
	unsigned int image_pixel_height;
	unsigned int image_bit_depth;
	unsigned int parent_object;
	unsigned short association_type;
	unsigned int association_description;
	unsigned int sequence_number;
	char *file_name;
	char *capture_date;
	char *modification_date;
	char *keywords;
};


/* Response to getstorageinfo - clause 5.5.3 pp 27 */

class mmptp_storageinfo
{
	public:
	mmptp_storageinfo() { memset(this, 0, sizeof(*this)); }
	~mmptp_storageinfo();
	int copy_contents(mmptp_container *container);
	int list_contents();

	unsigned short storage_type;
	unsigned short filesystem_type;
	unsigned short access_capability;
	unsigned int max_capacity_lsw;
	unsigned int max_capacity_msw;
	unsigned int free_space_bytes_lsw;
	unsigned int free_space_bytes_msw;
	unsigned int free_images;
	char *storage_description;
	char *volume_label;
};

#endif
#endif

